from __future__ import annotations

import copy

from kubernetes.client.models import V1LabelSelector, V1ObjectMeta, V1OwnerReference
from dragonk8s.lib.apimachinery.pkg import labels
from dragonk8s.lib.apimachinery.pkg.selection import Operator


NamespaceDefault = "default"
NamespaceAll = ""
NamespaceNone = ""
NamespaceSystem = "kube-system"
NamespacePublic = "kube-public"

StatusSuccess = "Success"
StatusFailure = "Failure"

LabelSelectorOpIn = "in"
LabelSelectorOpNotIn = "notin"
LabelSelectorOpExists = "exists"
LabelSelectorOpDoesNotExist = "doesnotexist"


def get_controller_of_no_copy(controllee) -> V1OwnerReference|None:
    refs = controllee.metadata.owner_references
    if not refs:
        return None
    for r in refs:
        if r.controller is True:
            return r
    return None


def get_controller_of(controllee) -> V1OwnerReference:
    res = get_controller_of_no_copy(controllee)
    return copy.copy(res)


def new_controller_ref(owner, gvk) -> V1OwnerReference:
    return V1OwnerReference(
        api_version=gvk.group_version,
        kind=gvk.kind,
        name=owner.metadata.name,
        uid=owner.metadata.uid,
        block_owner_deletion=True,
        controller=True,
    )


class TypeMeta(object):

    def __init__(self, kind="", api_version=""):
        self.kind = kind
        self.api_version = api_version


class ResourceVersionMatch(object):

    NotOlderThan = "NotOlderThan"
    Exact = "Exact"


class ListOptions(TypeMeta):

    def __init__(self, kind="", api_version="", label_selector="", field_selector="", watch=False,
                 allow_watch_bookmarks=False, resource_version="", resource_version_match=ResourceVersionMatch.Exact,
                 timeout_seconds=10, limit=10, _continue=""):
        super().__init__(kind, api_version)
        self.label_selector = label_selector
        self.field_selector = field_selector
        self.watch = watch
        self.allow_watch_bookmarks = allow_watch_bookmarks
        self.resource_version = resource_version
        self.resource_version_match = resource_version_match
        self.timeout_seconds = timeout_seconds
        self.limit = limit
        self._continue = _continue

    def to_params(self):
        return dict(
            label_selector=self.label_selector,
            field_selector=self.field_selector,
            resource_version=self.resource_version,
            timeout_seconds=self.timeout_seconds,
        )


class ListMeta(object):

    def __init__(self, self_link, resource_version, _continue, remaining_item_count):
        self.self_link = self_link
        self.resource_version = resource_version
        self._continue = _continue
        self.remaining_item_count = remaining_item_count


def labels_selector_as_selector(ps: V1LabelSelector) -> labels.Selector:
    if ps is None:
        return labels.nothing
    if ps.match_labels is None:
        ps.match_labels = {}
    if ps.match_expressions is None:
        ps.match_expressions = []
    if len(ps.match_labels) + len(ps.match_expressions) == 0:
        return labels.every_thing
    requirements = []
    for k, v in ps.match_labels.items():
        r = labels.new_requirement(k, Operator.Equals, [v])
        requirements.append(r)
    for expr in ps.match_expressions:
        op = None
        if expr.operator.lower() == LabelSelectorOpIn:
            op = Operator.In
        elif expr.operator.lower() == LabelSelectorOpNotIn:
            op = Operator.NotIn
        elif expr.operator.lower() == LabelSelectorOpExists:
            op = Operator.Exists
        elif expr.operator.lower() == LabelSelectorOpDoesNotExist:
            op = Operator.DoesNotExist
        else:
            raise Exception("%s is not a valid pod selector operator" % expr.operator)
        r = labels.new_requirement(expr.key, op, copy.copy(expr.values))
        requirements.append(r)
    selector = labels.new_selector()
    selector = selector.add(*requirements)
    return selector



import time
import unittest
from kubernetes import config
from threading import Event
from kubernetes.client.models.v1_object_meta import V1ObjectMeta
from dragonk8s.pkg.controller.controller_utils import RealPodControl
from dragonk8s.lib.client.tools import events
from kubernetes.client.models.v1_replica_set import V1ReplicaSet
from kubernetes.client.models.v1_pod_template_spec import V1PodTemplateSpec
from kubernetes.client.models.v1_pod_spec import V1PodSpec
from kubernetes.client.models.v1_container import V1Container
from kubernetes.client.models.v1_owner_reference import V1OwnerReference
from dragonk8s.dragon.api_client import ApiClient
from dragonk8s.lib.client import CommonClient
from dragonk8s.lib.apimeta import apigvk
from dragonk8s.dragon.configuration import Configuration


class RealPodControlTest(unittest.TestCase):

    def setUp(self):
        cfg = Configuration()
        config.load_kube_config(config_file="C:\\Users\\Administrator\\.kube\\config", client_configuration=cfg)
        Configuration.set_default(cfg)
        self.base_client = ApiClient()

        client = CommonClient(self.base_client)
        sink = events.EventSinkImpl(client)
        self.event_broadcaster = events.EventBroadcasterImpl(sink=sink)
        self.event_recorder = self.event_broadcaster.new_recorder("mhc-controller")
        self.pod_control = RealPodControl(client, self.event_recorder)
        self.controller = V1ReplicaSet(
            kind="ReplicaSet",
            api_version="v1",
            metadata=V1ObjectMeta(
                name="mhc-replica",
                namespace="default",
            ),
        )
        self.controller_ref = V1OwnerReference(
            name="mhc-replica",
            kind="ReplicaSet",
            api_version="v1",
            uid="123",
            controller=True,
            block_owner_deletion=True,

        )
        self.stop = Event()
        self.event_broadcaster.start_recording_to_sink(self.stop)
        self.event_broadcaster.start_structured_logging(self.stop)

    def test_create(self):
        ns = "default"
        container = V1Container(
            name="test",
            image="registry.cn-beijing.aliyuncs.com/mhc_base/linux-tools:net",
            command=["tail", "-f", "/dev/null"]
        )
        template = V1PodTemplateSpec(
            metadata=V1ObjectMeta(
                labels={"app": "test"}
            ),
            spec=V1PodSpec(
                containers=[container]
            )
        )
        self.pod_control.create_pod(ns, template, self.controller, self.controller_ref)

    def test_get(self):
        pod = self.pod_control.kube_client.get(namespace="default", name="mhc-replica-4rh58")
        print(pod.status.start_time.timestamp())

    def test_delete(self):
        self.pod_control.delete_pod("default", "mhc-replica-qsprs", self.controller)

    def test_patch(self):
        data = {
            "metadata": {
                "labels": {"app": "test", "aa2": "bb2"}
            }
        }
        self.pod_control.patch_pod("default", "mhc-replica-fhbmj", data)

    def test_patch2(self):
        data = {
            "metadata": {
                "uid": "",
                "ownerReferences": [
                    {
                        "apiVersion": "v1",
                        "kind": "ReplicaSet",
                        "name": "mmm2",
                        "uid": "222",
                        "controller": False,
                        "blockOwnerDeletion": True
                    }
                ],
                "finalizers": []
            }
        }
        self.pod_control.patch_pod("default", "mhc-replica-fhbmj", data)

    def tearDown(self) -> None:
        print("shut down")
        time.sleep(3)
        self.stop.set()
        self.event_broadcaster.shut_down()



import logging
import os
import sys
import threading
import signal
from dragonk8s.lib.client.informers import factory
from dragonk8s.lib.apimeta import apigvk
from dragonk8s.lib.client.informers.informer import get_informer
from dragonk8s.lib.client.informers.dragon.replicaset import ReplicaSetInformer
from kubernetes import config
from dragonk8s.lib.client import CommonClient, CommonWatch
from dragonk8s.dragon.api_client import ApiClient
from dragonk8s.dragon.configuration import Configuration
from dragonk8s.pkg.controller.replicaset.replica_set import ReplicaSetController
from dragonk8s.lib.pool import thread_manager

stop = threading.Event()
logging.basicConfig(level=logging.DEBUG, format="%(levelname)s %(thread)d %(filename)s-%(lineno)d(%(funcName)s): %(message)s")


def init_client() -> CommonClient:
    cfg = Configuration()
    if 'sys' in os.environ and os.environ['sys'] == "local":
        config.load_kube_config(config_file="C:\\Users\\Administrator\\.kube\\config", client_configuration=cfg)
    elif 'sys' in os.environ and os.environ['sys'] == "mac":
        config.load_kube_config(config_file="/Users/bytedance/mygit/deploy/0_config/my-k8s.conf", client_configuration=cfg)
    else:
        config.load_incluster_config(client_configuration=cfg)
    Configuration.set_default(cfg)
    base_client = ApiClient()
    client = CommonClient(base_client)
    return client


def exit_handler(signum, frame):
    logging.info("exit with signal.")
    if not stop.is_set():
        logging.info("set stop event.")
        stop.set()
        res = thread_manager.GlobalThreadManager.wait_stop_and_clean(10)
        if res:
            sys.exit(0)
        else:
            sys.exit(1)


def init_signal():
    signal.signal(signal.SIGINT, exit_handler)
    signal.signal(signal.SIGTERM, exit_handler)


def main():
    client = init_client()
    init_signal()
    f = factory.SharedInformerFactory(client, default_resync=0, stop=stop)
    pod_informer = get_informer(apigvk.CoreV1Pod, f)
    replicaset_informer = ReplicaSetInformer(informer_factory=f)
    replicaset_controller = ReplicaSetController(replicaset_informer, pod_informer, client, 300, stop)
    f.start(stop)
    f.wait_for_cache_sync(stop)
    replicaset_controller.run(1)


if __name__ == '__main__':
    main()



from threading import Lock, Event
from queue import Queue, Full, Empty
from dragonk8s.lib.apimachinery.pkg.watch.watch import Interface, Event as MyEvent
from dragonk8s.lib.pool.thread_manager import GlobalThreadManager

internalRunFunctionMarker = "internal-do-function"


class FullQueueBehavior(object):
    WaitIfChannelFull = 0
    DropIfChannelFull = 1


class Broadcaster(object):

    def __init__(self, queue_length: int, full_queue_behavior: int):
        self._watchers = {}
        self._next_watcher = 0
        self._incoming = Queue(25)
        self._watch_queue_length = queue_length
        self._full_queue_behavior = full_queue_behavior
        self._distributing = Event()
        self._stopped = False
        backend = GlobalThreadManager.new_thread(target=self._loop, generate_name="Broadcaster start", daemon=True)
        backend.start()

    def _block_queue(self, f):
        if self._stopped:
            return
        event = Event()

        def fake_obj():
            event.set()
            f()

        self._incoming.put(MyEvent(_type=internalRunFunctionMarker, obj=fake_obj))
        event.wait()

    def watch(self) -> Interface:
        watcher = BroadcasterWatcher()

        def f():
            cur_id = self._next_watcher
            self._next_watcher += 1
            watcher.id = cur_id
            watcher.m = self
            watcher.result = Queue(self._watch_queue_length)
            self._watchers[cur_id] = watcher

        self._block_queue(f)
        if watcher.m is None:
            raise Exception("broadcaster already stopped")
        return watcher

    def watch_with_prefix(self, queued_events) -> Interface:
        watcher = BroadcasterWatcher()

        def f():
            cur_id = self._next_watcher
            self._next_watcher += 1
            length = self._watch_queue_length
            n = len(queued_events) + 1
            if n > length:
                length = n
            watcher.id = cur_id
            watcher.m = self
            watcher.result = Queue(length)
            self._watchers[cur_id] = watcher
            for queued_event in queued_events:
                watcher.result.put(queued_event)
        self._block_queue(f)
        if watcher.m is None:
            raise Exception("broadcaster already stopped")
        return watcher

    def stop_watching(self, _id: int):

        def f():
            if _id in self._watchers.keys():
                del self._watchers[_id]
        self._block_queue(f)

    def _close_all(self):
        self._watchers.clear()

    def action(self, action: str, obj):
        if not self._stopped:
            try:
                self._incoming.put(MyEvent(action, obj), block=True, timeout=5)
            except Full:
                return

    def action_or_drop(self, action: str, obj) -> bool:
        try:
            self._incoming.put(MyEvent(action, obj), block=False)
            return True
        except Full:
            return False

    def shutdown(self):

        def f():
            self._stopped = True
        self._block_queue(f)
        self._distributing.wait()

    def _loop(self):
        while not self._stopped:
            try:
                event = self._incoming.get(block=True, timeout=5)
                if event.type == internalRunFunctionMarker:
                    event.obj()
                    continue
                self._distribute(event)
            except Empty:
                continue
        self._close_all()
        self._distributing.set()

    def _distribute(self, event):
        if self._full_queue_behavior == FullQueueBehavior.DropIfChannelFull:
            for w in self._watchers.values():
                if not w.stopped:
                    try:
                        w.result.put(event, block=False)
                    except Full:
                        break
        else:
            for w in self._watchers.values():
                while not w.stopped:
                    try:
                        w.result.put(event, block=True, timeout=5)
                        break
                    except Full:
                        continue


class BroadcasterWatcher(Interface):

    def __init__(self, _id=0, result_queue: Queue = None, m: Broadcaster = None):
        self.result = result_queue
        self.stopped = False
        self._stop_lock = Lock()
        self.id = _id
        self.m = m

    def stop(self):
        if self.stopped:
            return
        if not self._stop_lock.acquire():
            return
        try:
            if self.stopped:
                return
            self.stopped = True
            self.m.stop_watching(self.id)
        finally:
            self._stop_lock.release()

    def get_result_queue(self) -> Queue:
        return self.result



import threading
import traceback
import json
from queue import Empty
import logging
import time
from kubernetes.client.models.v1_object_meta import V1ObjectMeta
from kubernetes.client.models.v1_event_source import V1EventSource
from kubernetes.client.models.v1_object_reference import V1ObjectReference
from dragonk8s.lib.apimachinery.pkg.watch import watch, mux
from kubernetes.client.models.core_v1_event import CoreV1Event
from kubernetes.client.exceptions import ApiException
from dragonk8s.lib.apimachinery.pkg.api import errors
from copy import deepcopy
from threading import RLock
from dragonk8s.lib.util.cache import lru
from dragonk8s.lib.client.util.flow_control import TokenBucketRateLimiter
from dragonk8s.lib.util import timeutil
from dragonk8s.lib.apimachinery.pkg.apis import meta_v1
from dragonk8s.lib.client.tools import reference
from dragonk8s.lib.pool.thread_manager import GlobalThreadManager


EventTypeNormal = "Normal"
EventTypeWarning = "Warning"


def validate_event_type(eventtype: str) -> bool:
    if eventtype == EventTypeNormal or eventtype == EventTypeWarning:
        return True
    return False


class RecordException(Exception):
    pass


def is_key_not_found_error(ex) -> bool:
    return ex is not None and ex.status == 404


def event_aggregator_by_reason_func(event: CoreV1Event) -> (str, str):
    ss = [event.source.component, event.source.host, event.involved_object.kind, event.involved_object.namespace,
          event.involved_object.name, event.involved_object.uid, event.involved_object.api_version,
          event.type, event.reason, event.reporting_component, event.reporting_instance]
    tos = [str(s) for s in ss]
    return "".join(tos), event.message


def event_aggregator_by_reason_message_func(event: CoreV1Event) -> str:
    return "(combined from similar events): %s" % event.message


def get_spam_key(event: CoreV1Event) -> str:
    ss = [event.source.component, event.source.host, event.involved_object.kind, event.involved_object.namespace,
          event.involved_object.name, event.involved_object.uid, event.involved_object.api_version]
    tos = [str(s) for s in ss]
    return "".join(tos)


def get_event_key(event: CoreV1Event) -> str:
    ss = [event.source.component, event.source.host, event.involved_object.kind, event.involved_object.namespace,
          event.involved_object.name, event.involved_object.field_path, event.involved_object.uid,
          event.involved_object.api_version, event.type, event.reason, event.message]
    tos = [str(s) for s in ss]
    return "".join(tos)


class EventSink(object):

    def create(self, event: CoreV1Event) -> CoreV1Event:
        return CoreV1Event()

    def update(self, event: CoreV1Event) -> CoreV1Event:
        return CoreV1Event()

    def patch(self, old_event: CoreV1Event, data) -> CoreV1Event:
        return CoreV1Event()


class EventRecorder(object):

    def event(self, obj, event_type: str, reason: str, message: str):
        pass

    def eventf(self, obj, eventtype: str, reason: str, message_fmt: str, *args):
        pass

    def annotated_eventf(self, obj, annotations: dict, eventtype: str, reason: str, message_fmt: str, *args):
        pass


class EventBroadcaster(object):

    def start_event_watcher(self, event_handler) -> watch.Interface:
        pass

    def start_recording_to_sink(self, sink: EventSink) -> watch.Interface:
        pass

    def start_logging(self, logf) -> watch.Interface:
        pass

    def start_structured_logging(self, level) -> watch.Interface:
        pass

    def new_recorder(self, scheme, source: V1EventSource) -> EventRecorder:
        pass

    def shut_down(self):
        pass


class SpamRecord(object):

    def __init__(self, rate_limiter: TokenBucketRateLimiter = None):
        self.rate_limiter = rate_limiter


class EventSourceObjectSpamFilter(object):

    def __init__(self, lru_cache_size: int, burst: int, qps: float, spam_key_func):
        self._lock = RLock()
        self._cache = lru.Cache(lru_cache_size)
        self._burst = burst
        self._qps = qps
        self._spam_key_func = spam_key_func

    def filter(self, event: CoreV1Event):
        record = SpamRecord()
        event_key = self._spam_key_func(event)
        try:
            self._lock.acquire()
            value, found = self._cache.get(event_key)
            if found:
                record = value
            if record.rate_limiter is None:
                record.rate_limiter = TokenBucketRateLimiter(self._qps, self._burst)
            fil = not record.rate_limiter.try_accept()
            self._cache.add(event_key, record)
            return fil
        finally:
            self._lock.release()


class AggregateRecord(object):

    def __init__(self, local_keys=None, last_timestamp=0):
        if local_keys is None:
            self.local_keys = []
        else:
            self.local_keys = local_keys
        self.last_timestamp = last_timestamp


class EventAggregator(object):

    def __init__(self, lru_cache_size, key_func, message_func, max_events, max_interval_in_seconds):
        self._lock = RLock()
        self._cache = lru.Cache(lru_cache_size)
        self._key_func = key_func
        self._message_func = message_func
        self._max_events = max_events
        self._max_interval_in_seconds = max_interval_in_seconds

    def event_aggregate(self, new_event: CoreV1Event) -> (CoreV1Event, str):
        now = time.time()
        record = AggregateRecord()
        event_key = get_event_key(new_event)
        aggregate_key, local_key = self._key_func(new_event)

        try:
            self._lock.acquire(True)
            value, found = self._cache.get(aggregate_key)
            if found:
                record = value

            max_interval = self._max_interval_in_seconds
            interval = now - record.last_timestamp
            if interval > max_interval:
                record = AggregateRecord(local_keys=[])

            record.local_keys.append(local_key)
            record.last_timestamp = now
            self._cache.add(aggregate_key, record)

            if len(record.local_keys) < self._max_events:
                return new_event, event_key

            record.local_keys.pop()
            event = CoreV1Event(
                api_version="v1",
                kind="Event",
                metadata=V1ObjectMeta(
                    name="%s.%d" % (new_event.involved_object.name, now),
                    namespace=new_event.metadata.namespace
                ),
                count=1,
                first_timestamp=timeutil.to_time_str(now),
                involved_object=new_event.involved_object,
                last_timestamp=timeutil.to_time_str(now),
                message=self._message_func(new_event),
                type=new_event.type,
                reason=new_event.reason,
                source=new_event.source,
            )
            return event, aggregate_key
        finally:
            self._lock.release()


class EventLog(object):

    def __init__(self, count: int = 0, first_timestamp: float=0, name="", resource_version=""):
        self.count = count
        self.first_timestamp = first_timestamp
        self.name = name
        self.resource_version = resource_version


class EventLogger(object):

    def __init__(self, lru_cache_entries: int):
        self._lock = RLock()
        self._cache = lru.Cache(lru_cache_entries)

    def event_observe(self, new_event: CoreV1Event, key: str) -> (CoreV1Event, str):
        event = deepcopy(new_event)
        patch = {}
        try:
            self._lock.acquire(True)
            last_observation = self._last_event_observation_from_cache(key)
            if last_observation.count > 0:
                event.name = last_observation.name
                event.metadata.resource_version = last_observation.resource_version
                event.first_timestamp = timeutil.to_time_str(last_observation.first_timestamp)
                event.count = last_observation.count + 1

                patch = dict(
                    count=event.count,
                    lastTimestamp=timeutil.to_time_str(event.last_timestamp),
                    message=event.message
                )
            self._cache.add(key, EventLog(count=event.count,
                                          first_timestamp=timeutil.parse_time(event.first_timestamp),
                                          name=event.metadata.name,
                                          resource_version=event.metadata.resource_version))
            return event, patch

        finally:
            self._lock.release()

    def _last_event_observation_from_cache(self, key) -> EventLog:
        value, ok = self._cache.get(key)
        if ok:
            return value
        return EventLog()

    def update_state(self, event: CoreV1Event):
        key = get_event_key(event)
        try:
            self._lock.acquire(True)
            self._cache.add(key, EventLog(count=event.count,
                                          first_timestamp=timeutil.parse_time(event.first_timestamp),
                                          name=event.metadata.name,
                                          resource_version=event.metadata.resource_version))
        finally:
            self._lock.release()


class CorrelatorOptions(object):

    def __init__(self, lru_cache_size=4096, burst_size=25, qps=float(1./300.),
                 key_func=event_aggregator_by_reason_func,
                 message_func=event_aggregator_by_reason_message_func,
                 max_events=10, max_interval_in_seconds=600, clock=None,
                 spam_key_func=get_spam_key):
        self.lru_cache_size = lru_cache_size
        self.BurstSize = burst_size
        self.qps = qps
        self.key_func = key_func
        self.message_func = message_func
        self.max_events = max_events
        self.max_interval_in_seconds = max_interval_in_seconds
        self.clock = clock
        self.spam_key_func = spam_key_func


class EventCorrelateResult(object):

    def __init__(self, event=None, patch=None, skip=False):
        self.event = event
        self.patch = patch
        self.skip = skip


class EventCorrelator(object):

    def __init__(self):
        cache_size = 4096
        spam_filter = EventSourceObjectSpamFilter(cache_size, 25, float(1.0/300.0), get_spam_key)
        self.filter_func = spam_filter.filter
        self.aggregator = EventAggregator(cache_size,
                                          event_aggregator_by_reason_func,
                                          event_aggregator_by_reason_message_func,
                                          10,
                                          600)
        self.logger = EventLogger(cache_size)

    @classmethod
    def new_with_options(cls, options: CorrelatorOptions):
        spam_filter = EventSourceObjectSpamFilter(options.lru_cache_size,
                                                  options.BurstSize,
                                                  options.qps,
                                                  options.spam_key_func)
        res = cls()
        res.filter_func = spam_filter.filter
        res.aggregator = EventAggregator(options.lru_cache_size,
                                         options.key_func,
                                         options.message_func,
                                         options.max_events,
                                         options.max_interval_in_seconds)
        res.logger = EventLogger(options.lru_cache_size)
        return res

    def event_correlate(self, new_event: CoreV1Event) -> EventCorrelateResult:
        if new_event is None:
            raise RecordException("event is None")
        aggregate_event, c_key = self.aggregator.event_aggregate(new_event)
        observed_event, patch = self.logger.event_observe(aggregate_event, c_key)
        if self.filter_func(observed_event):
            return EventCorrelateResult(skip=True)
        return EventCorrelateResult(event=observed_event, patch=patch)

    def update_state(self, event: CoreV1Event):
        self.logger.update_state(event=event)


def _record_event(sink: EventSink, event: CoreV1Event, patch: dict,
                  update_existing_event: bool, event_correlator: EventCorrelator) -> bool:
    new_event = None
    create = False
    if update_existing_event:
        try:
            new_event = sink.patch(event, patch)
        except ApiException as e:
            if is_key_not_found_error(e):
                create = True
            else:
                logging.error("server rejected event: %s: %s (will not retry)" % (str(event), e))
                return True
        except Exception as e:
            logging.error("patch event(%s) error: %s (may retry after sleeping)" % (str(event), e))
            return False
        else:
            event_correlator.update_state(new_event)
            return True
    if not update_existing_event or create:
        event.metadata.resource_version = ''
        try:
            new_event = sink.create(event)
        except ApiException as e:
            if errors.is_already_exists(e):
                event_correlator.update_state(new_event)
                return True
            else:
                logging.error("server rejected event: %s: %s (will not retry)" % (str(event), e))
                return True
        except Exception as e:
            logging.error("patch event(%s) error: %s (may retry after sleeping)" % (str(event), e))
            return False
        else:
            event_correlator.update_state(new_event)
            return True
    return False


def _record_to_sink(sink: EventSink, event: CoreV1Event, event_correlator: EventCorrelator, sleep_duration: int):
    event_copy = deepcopy(event)
    result = None
    try:
        result = event_correlator.event_correlate(event_copy)
    except Exception as e:
        traceback.format_exc()
        logging.error("event_correlate error: %s" % e)
        return
    if result.skip:
        return
    tries = 0
    while True:
        if _record_event(sink, result.event, result.patch, result.event.count > 1, event_correlator):
            break
        tries += 1
        if tries >= 12:
            logging.error("unable to write event %s (retry limit exceeded)" % str(event_copy.__dict__))
            break
        time.sleep(sleep_duration)


class EventBroadcasterImpl(EventBroadcaster):

    def __init__(self, options: CorrelatorOptions = None, stop=threading.Event()):
        super().__init__()
        if options is None:
            options = CorrelatorOptions()
        self.watcher = mux.Broadcaster(1000, mux.FullQueueBehavior.DropIfChannelFull)
        self.sleep_duration = 10
        self.options = options
        self.stop = stop

    def start_recording_to_sink(self, sink: EventSink) -> watch.Interface:
        event_correlator = EventCorrelator.new_with_options(self.options)

        def event_handler(event: CoreV1Event):
            _record_to_sink(sink, event, event_correlator, self.sleep_duration)
        return self.start_event_watcher(event_handler)

    def shut_down(self):
        self.watcher.shutdown()

    def start_event_watcher(self, event_handler) -> watch.Interface:
        watcher = self.watcher.watch()

        def handle():
            try:
                queue = watcher.get_result_queue()
                while not self.stop.is_set():
                    try:
                        watch_event = queue.get(True, 5)
                    except Empty:
                        continue
                    else:
                        event_handler(watch_event.obj)
                    # todo
            except Exception as e:
                logging.error("handle event error: %s" % e)

        backend = GlobalThreadManager.new_thread(target=handle, generate_name="Broadcaster Loop")
        backend.start()
        return watcher

    def start_logging(self, logf) -> watch.Interface:
        def handle(e: CoreV1Event):
            logf("event(%s): type: %s, reason: %s,  %s" % (str(e.involved_object.__dict__), e.type, e.reason, e.message))
        return self.start_event_watcher(handle)

    def new_recorder(self, source: V1EventSource, scheme=None) -> EventRecorder:
        return RecorderImpl(scheme, source, self.watcher)

    def start_structured_logging(self, level=0) -> watch.Interface:
        def handle(e: CoreV1Event):
            logging.info("event: %s" % str(e.__dict__))
        return self.start_event_watcher(handle)


class RecorderImpl(EventRecorder):

    def __init__(self, scheme, source: V1EventSource, watcher: mux.Broadcaster):
        super().__init__()
        self.scheme = scheme
        self.source = source
        self.watcher = watcher

    def _make_event(self, ref: V1ObjectReference, annotations: dict, eventtype: str, reason: str, message: str) -> CoreV1Event:
        now = time.time()
        time_str = timeutil.to_time_str(now)
        namespace = ref.namespace
        if namespace == "":
            namespace = meta_v1.NamespaceDefault
        return CoreV1Event(
            api_version="v1",
            kind="Event",
            metadata=V1ObjectMeta(
                name="%s.%d" %(ref.name, now),
                namespace=namespace,
                annotations=annotations,
            ),
            involved_object=ref,
            type=eventtype,
            reason=reason,
            message=message,
            first_timestamp=time_str,
            last_timestamp=time_str,
            # event_time=time_str,
            count=1,
        )

    def _generate_event(self, obj, annotations: dict, eventtype: str, reason: str, message: str):
        try:
            ref = reference.get_reference(obj)
        except Exception as e:
            logging.error("Could not construct reference to %s due to %s, will not report event: '%s' '%s' '%s'"
                          % (str(obj), str(e), eventtype, reason, message))
            return
        if not validate_event_type(eventtype):
            logging.error("Unsupported event type: %s" % eventtype)
            return

        event = self._make_event(ref, annotations, eventtype, reason, message)
        event.source = self.source
        sent = self.watcher.action_or_drop(watch.EventType.Added, event)
        if not sent:
            logging.error("unable to record evnet: too many queued events, dropped event: %s", str(event))

    def event(self, obj, eventtype: str, reason: str, message: str):
        self._generate_event(obj, None, eventtype, reason, message)

    def eventf(self, obj, eventtype: str, reason: str, message_fmt: str, *args):
        self._generate_event(obj, None, eventtype, reason, message_fmt.format(*args))

    def annotated_eventf(self, obj, annotations: dict, eventtype: str, reason: str, message_fmt: str, *args):
        self._generate_event(obj, annotations, eventtype, reason, message_fmt.format(*args))



import unittest
import time
from dragonk8s.lib.util import timeutil
import queue


class UtilTest(unittest.TestCase):

    def test1(self):
        now = 1671851469.805154
        self.assertEqual('2022-12-24T11:11:09.805154Z', timeutil.to_time_str_with_ns(now))

    def test2(self):
        q = queue.Queue(2)
        q.put("a",  True, 5)
        print(q.qsize())
        print(q.get(True, 5))
        print(q.qsize())
        print(q.get(True, 5))


import json
import logging
import threading
from dragonk8s.lib.apimachinery.pkg.util import wait


class KeyLister(object):

    def list_keys(self) -> list:
        return []


class KeyGetter(object):

    def get_by_key(self, key: str) -> (object, bool):
        return None, False


class KeyListerGetter(KeyLister, KeyGetter):
    pass


class Store(KeyListerGetter):

    def add(self, obj):
        pass

    def update(self, obj):
        pass

    def delete(self, obj):
        pass

    def list(self):
        pass

    def list_keys(self):
        pass

    def get(self, obj):
        pass

    def get_by_key(self, key):
        pass

    def replace(self, data, resource_version):
        pass

    def resync(self):
        pass


class Indexer(Store):
    def index(self, index_name, obj):
        pass

    def index_keys(self, index_name, indexed_value):
        pass

    def list_index_func_values(self, index_name):
        pass

    def by_index(self, index_name, indexed_value):
        pass

    def get_indexers(self):
        pass

    def add_indexers(self, new_indexers):
        pass


class ThreadSafeStore(object):

    def add(self, key, obj):
        pass

    def update(self, key, obj):
        pass

    def delete(self, key, obj):
        pass

    def get(self, key):
        pass

    def list(self):
        pass

    def list_keys(self):
        pass

    def replace(self, items: dict, resource_version: str):
        pass

    def index(self, index_name, obj):
        pass

    def index_keys(self, index_name, indexed_value):
        pass

    def list_index_func_values(self, index_name):
        pass

    def by_index(self, index_name, indexed_value):
        pass

    def get_indexers(self):
        pass

    def add_indexers(self, new_indexers):
        pass

    def resync(self):
        pass


def meta_namespace_key_func(obj):
    if isinstance(obj, str):
        return obj
    if hasattr(obj, "metadata"):
        meta = obj.metadata
        if len(meta.namespace) > 0:
            return "{}/{}".format(meta.namespace, meta.name)
        return meta.name
    raise Exception("object has no meta")


def split_meta_namespace_key(key: str):
    parts = key.split("/")
    l = len(parts)
    if l == 1:
        return "", parts[0]
    if l == 2:
        return parts[0], parts[1]
    raise Exception("unexpected key format: %s" % key)


class ThreadSafeMap(ThreadSafeStore):

    def __init__(self, indexers: dict, indices: dict):
        super().__init__()
        self._lock = threading.Lock()
        self._items = dict()
        self._indexers = indexers
        self._indices = indices

    def _delete_key_from_index(self, key: str, index_value: str, index: dict):
        if index_value not in index:
            return
        s = index[index_value]
        if key in s:
            s.remove(key)
        if len(s) == 0:
            del index[index_value]

    def add_key_to_index(self, key: str, index_value: str, index: dict):
        if index_value not in index:
            index[index_value] = set()
        index[index_value].add(key)

    def _update_indices(self, old_obj, new_obj, key: str):
        old_index_values = []
        index_values = []
        for name, index_func in self._indexers.items():
            if old_obj is not None:
                old_index_values = index_func(old_obj)
            else:
                old_index_values.clear()
            if new_obj is not None:
                index_values = index_func(new_obj)
            else:
                index_values.clear()

            if name not in self._indices:
                self._indices[name] = dict()

            index = self._indices[name]

            if len(index_values) == 1 and len(old_index_values) == 1 and index_values[0] == old_index_values[0]:
                continue

            for value in old_index_values:
                self._delete_key_from_index(key, value, index)

            for value in index_values:
                self.add_key_to_index(key, value, index)

    def add(self, key, obj):
        self.update(key, obj)

    def update(self, key, obj):
        try:
            self._lock.acquire(True)
            old_obj = None
            if key in self._items:
                old_obj = self._items[key]
            self._items[key] = obj
            self._update_indices(old_obj, obj, key)
        finally:
            self._lock.release()

    def delete(self, key, obj):
        try:
            self._lock.acquire(True)
            if key in self._items:
                self._update_indices(obj, None, key)
                del self._items[key]
        finally:
            self._lock.release()

    def get(self, key):
        try:
            self._lock.acquire(True)
            if key in self._items:
                return self._items[key], True
            return None, False
        finally:
            self._lock.release()

    def list(self):
        try:
            self._lock.acquire(True)
            return self._items.values()
        finally:
            self._lock.release()

    def list_keys(self):
        with self._lock:
            return self._items.keys()

    def replace(self, items: dict, resource_version: str):
        try:
            self._lock.acquire(True)
            self._items = items
            self._indices = dict()
            for key, item in self._items.items():
                self._update_indices(None, item, key)
        finally:
            self._lock.release()

    def _get_index_func(self, index_name):
        if index_name not in self._indexers:
            raise Exception("Index with name %s does not exist" % index_name)
        return self._indexers[index_name]

    def index(self, index_name, obj):
        try:
            self._lock.acquire(True)
            index_func = self._get_index_func(index_name)
            indexed_values = index_func(obj)
            index = self._indices[index_name]
            store_key_set = {}
            if len(indexed_values) == 1:
                store_key_set = index[indexed_values[0]]
            else:
                for indexed_value in indexed_values:
                    for key in index[indexed_value]:
                        store_key_set.add(key)
            return [self._items[store_key] for store_key in store_key_set]
        finally:
            self._lock.release()

    def index_keys(self, index_name, indexed_value):
        try:
            self._lock.acquire(True)
            index_func = self._get_index_func(index_name)
            index = self._indices[index_name]
            if indexed_value not in index:
                return []
            s = index[indexed_value]
            return list(s)
        finally:
            self._lock.release()

    def list_index_func_values(self, index_name):
        try:
            self._lock.acquire(True)
            if index_name not in self._indices:
                return []
            index = self._indices[index_name]
            return index.keys()
        finally:
            self._lock.release()

    def by_index(self, index_name, indexed_value):
        try:
            self._lock.acquire(True)
            index_func = self._get_index_func(index_name)
            index = self._indices[index_name]
            if indexed_value not in index:
                return []
            s = index[indexed_value]
            return [self._items[key] for key in s]
        finally:
            self._lock.release()

    def get_indexers(self):
        return self._indexers

    def add_indexers(self, new_indexers):
        try:
            self._lock.acquire(True)
            if len(self._items) > 0:
                raise Exception("cannot add indexers to running index")
            old_keys = self._indexers.keys()
            new_keys = new_indexers.keys()
            for new_key in new_keys:
                if new_key in old_keys:
                    raise Exception("indexer conflict: %s" % new_key)

            for k, v in new_indexers.items():
                self._indexers[k] = v

        finally:
            self._lock.release()

    def resync(self):
        # nothing to do
        pass


class Cache(Indexer):

    def __init__(self, key_func, indexers: dict):
        super().__init__()
        self._key_func = key_func
        self._cache_storage = ThreadSafeMap(indexers, {})

    @classmethod
    def new_Store(cls, key_func):
        return cls(key_func, {})

    @classmethod
    def new_indexer(cls, key_func, indexers: dict):
        return cls(key_func, indexers)

    def add(self, obj):
        key = self._key_func(obj)
        self._cache_storage.add(key, obj)

    def update(self, obj):
        key = self._key_func(obj)
        self._cache_storage.update(key, obj)

    def delete(self, obj):
        key = self._key_func(obj)
        self._cache_storage.delete(key)

    def list(self):
        return self._cache_storage.list()

    def list_keys(self):
        return self._cache_storage.list_keys()

    def get(self, obj):
        key = self._key_func(obj)
        return self.get_by_key(key)

    def get_by_key(self, key):
        return self._cache_storage.get(key)

    def replace(self, data, resource_version):
        items = {}
        for item in data:
            key = self._key_func(item)
            items[key] = item
        self._cache_storage.replace(items, resource_version)

    def resync(self):
        pass

    def index(self, index_name, obj):
        return self._cache_storage.index(index_name, obj)

    def index_keys(self, index_name, indexed_value):
        return self._cache_storage.index_keys(index_name, indexed_value)

    def list_index_func_values(self, index_name):
        return self._cache_storage.list_index_func_values(index_name)

    def by_index(self, index_name, indexed_value):
        self._cache_storage.by_index(index_name, indexed_value)

    def get_indexers(self):
        return self._cache_storage.get_indexers()

    def add_indexers(self, new_indexers):
        self._cache_storage.add_indexers(new_indexers)


class DeletedFinalStateUnknown(object):

    def __init__(self, key, obj=None):
        self.key = key
        self.obj = obj


class KeyError(Exception):

    def __init__(self, obj, e: Exception):
        self.e = e
        self.obj = obj
        objstr = json.dumps(obj)
        super(KeyError, self).__init__("couldn't create key for object: %s: %s", objstr, str(e))

    def unwrap(self) -> Exception:
        return self.e


class MutationDetector(object):

    def add_object(self, obj):
        pass

    def run(self, stop: threading.Event):
        pass


class DummyMutationDetector(MutationDetector):
    pass


def wait_for_cache_sync(stop: threading.Event, *cache_syncs) -> bool:
    def sync():
        for s in cache_syncs:
            if not s():
                return False
        return True
    try:
        ok = wait.poll(sync, 1, 0, stop, immadiate=True)
        # todo
    except Exception as e:
        logging.error("wait poll failed: %s" % str(e))
        return False
    if not ok:
        logging.info("caches no populated")
        return False
    logging.info("caches populated")
    return True


def wait_for_named_cache_sync(controller_name: str, stop: threading.Event, *cache_syncs) -> bool:
    logging.info("Waiting for caches to sync for %s" % controller_name)
    if not wait_for_cache_sync(stop, *cache_syncs):
        logging.error("unable to sync caches for %s" % controller_name)
        return False
    logging.info("Caches are synced for %s" % controller_name)
    return True



NamespaceIndex = "namespace"


def meta_namespace_index_func(obj) -> list:
    if hasattr(obj, "namespace"):
        metadata = obj
    else:
        metadata = obj.metadata
    return [metadata.namespace]



from queue import Queue


class Interface(object):

    def stop(self):
        pass

    def get_result_queue(self) -> Queue:
        pass


class EventType(object):
    Added = "ADDED"
    Modified = "MODIFIED"
    Deleted = "DELETED"
    Bookmark = "BOOKMARK"
    Error = "ERROR"
    End = "End"


class Event(object):

    def __init__(self, _type: str = EventType.Added, obj=None):
        self.type = _type
        self.obj = obj



import copy
import logging

from dragonk8s.lib.apimachinery.pkg import selection
from dragonk8s.lib.apimachinery.pkg.util.validation import field
from dragonk8s.lib.apimachinery.pkg.util import validation
from dragonk8s.lib.apimachinery.pkg.selection import Operator
from dragonk8s.lib.util import string

_unary_operators = [Operator.Exists, Operator.DoesNotExist]
_binary_operators = [Operator.In, Operator.NotIn, Operator.Equals, Operator.DoubleEquals, Operator.NotEquals,
                     Operator.GreaterThan, Operator.LessThan]
valid_requirement_operators = copy.copy(_unary_operators)
valid_requirement_operators.extend(_binary_operators)


class Labels(object):

    def has(self, label: str) -> bool:
        pass

    def get(self, label: str) -> str:
        pass


def _validate_label_key(k: str, path: field.Path):
    try:
        validation.is_qualified_name(k)
    except Exception as e:
        raise field.invalid(path, k, str(e))


def _validate_label_value(k, v, path: field.Path):
    try:
        validation.is_valid_label_value(v)
    except Exception as e:
        raise field.invalid(path.key(k), v, str(e))


class Requirement(object):

    def __init__(self, key: str, op: str, vals: list):
        self.key = key
        self.operator = op
        self.str_values = vals

    def has_value(self, value: str) -> bool:
        return value in self.str_values

    def matches(self, ls: Labels):
        if self.operator in {Operator.In, Operator.Equals, Operator.DoubleEquals}:
            if not ls.has(self.key):
                return False
            return self.has_value(ls.get(self.key))
        elif self.operator in {Operator.NotIn, Operator.NotEquals}:
            if not ls.has(self.key):
                return True
            return not self.has_value(ls.get(self.key))
        elif self.operator == Operator.Exists:
            return ls.has(self.key)
        elif self.operator == Operator.DoesNotExist:
            return not ls.has(self.key)
        elif self.operator in {Operator.GreaterThan, Operator.LessThan}:
            if not ls.has(self.key):
                return False
            try:
                ls_value = int(ls.get(self.key))
            except Exception as e:
                logging.warning("parse failed for value: %s in label %s: %s", str(ls.get(self.key)), ls, str(e))
                return False
            if len(self.str_values) != 1:
                logging.warning("Invalid value count %s of requirement %s, "
                                "for 'Gt', 'Lt' operators, exactly one value is required", len(self.str_values), self)
                return False
            for v in self.str_values:
                try:
                    self_v = int(v)
                except Exception as e:
                    logging.warning("parse failed for value %s in requirement %s,"
                                    " for 'Gt', 'Lt' operators, the value must be an integer", v, self)
                    return False
            if self.operator == Operator.GreaterThan:
                return ls_value > self_v
            elif self.operator == Operator.LessThan:
                return ls_value < self_v
        else:
            return False

    def __repr__(self):
        res = ""
        if self.operator == Operator.DoesNotExist:
            res += "!"
        res += self.key
        if self.operator == Operator.Equals:
            res += "="
        elif self.operator == Operator.DoubleEquals:
            res += "=="
        elif self.operator == Operator.NotEquals:
            res += "!="
        elif self.operator == Operator.In:
            res += " in "
        elif self.operator == Operator.NotIn:
            res += " notin "
        elif self.operator == Operator.GreaterThan:
            res += ">"
        elif self.operator == Operator.LessThan:
            res += "<"
        elif self.operator in {Operator.Exists, Operator.DoesNotExist}:
            return res

        if self.operator in {Operator.In, Operator.NotIn}:
            res += "("
        if len(self.str_values) == 1:
            res += self.str_values[0]
        else:
            res += ", ".join(self.str_values)
        if self.operator in {Operator.In, Operator.NotIn}:
            res += ")"
        return res

    def __str__(self):
        return self.__repr__()

    @property
    def values(self):
        return self.str_values

    def __eq__(self, other):
        if self.key != other.key:
            return False
        if self.operator != other.operator:
            return False
        return string.list_equal(self.str_values, other.str_values)


def new_requirement(key: str, op: str, vals: list, *opts):
    path = field.to_path(opts)
    _validate_label_key(key, path.child("key"))

    value_path = path.child("values")
    if op in {Operator.In, Operator.NotIn}:
        if len(vals) == 0:
            raise field.invalid(value_path, vals, "for 'in', 'not in' operators, values set can't be empty")
    elif op in {Operator.In, Operator.DoubleEquals, Operator.NotEquals}:
        if len(vals) != 1:
            raise field.invalid(value_path, vals, "exact-match compatibility requires on single value")
    elif op in {Operator.Exists, Operator.DoesNotExist}:
        if len(vals) != 0:
            raise field.invalid(value_path, vals, "values set must be empty for exists and does not exist")
    elif op in {Operator.GreaterThan, Operator.LessThan}:
        if len(vals) != 1:
            raise field.invalid(value_path, vals, "for 'Gt', 'Lt' operators, exactly one value is required")
        for i in range(len(vals)):
            try:
                int(vals[i])
            except:
                raise field.invalid(value_path.index(i), vals[i],
                                    "for 'Gt', 'Lt' operators, the value mutst be an interger")
    else:
        raise field.not_supported(path.child("operator"), op, valid_requirement_operators)
    for i in range(len(vals)):
        _validate_label_value(key, vals[i], value_path.index(i))

    return Requirement(key=key, op=op, vals=vals)


class Set(Labels):

    def __init__(self, data=None):
        self._data = {} if not data else data

    def __repr__(self):
        selector = []
        for k, v in self._data.items():
            selector.append("{}={}".format(k, v))
        return ",".join(selector)

    def __str__(self):
        return self.__repr__()

    def has(self, label: str) -> bool:
        return label in self._data

    def get(self, label: str) -> str:
        return self._data[label] if label in self._data else None


class Selector(object):

    def matches(self, labels: Labels) -> bool:
        pass

    def empty(self) -> bool:
        pass

    def add(self, *r: Requirement):
        pass

    def requirements(self) -> (list, bool):
        pass

    def requires_exact_match(self, label: str) -> (str, bool):
        pass


class _InternalSelector(Selector):

    def __init__(self, data=None):
        if not data:
            self._data = []
        else:
            self._data = data

    def matches(self, labels: Labels) -> bool:
        for r in self._data:
            if not r.matches(labels):
                return False
        return True

    def empty(self) -> bool:
        return len(self._data) == 0

    def add(self, *r: Requirement):
        self._data.extend(r)
        # todo sort
        return self

    def requirements(self) -> (list, bool):
        return self._data, True

    def requires_exact_match(self, label: str) -> (str, bool):
        for r in self._data:
            if r.key == label:
                if r.operator in {Operator.Equals, Operator.DoubleEquals, Operator.In}:
                    if len(r.str_values) == 1:
                        return r.str_values[0], True
                return "", False
        return "", False

    def __repr__(self):
        res = []
        for s in self._data:
            res.append(str(s))
        return ",".join(res)

    def __str__(self):
        return self.__repr__()


class _NothingSelector(Selector):
    def matches(self, labels: Labels) -> bool:
        return False

    def empty(self) -> bool:
        return False

    def add(self, *r: Requirement):
        pass

    def requirements(self) -> (list, bool):
        return [], False

    def requires_exact_match(self, label: str) -> (str, bool):
        return "", False

    def __repr__(self):
        return ""

    def __str__(self):
        return ""


def selector_from_validated_set(ls: Set) -> Selector:
    if ls is None or len(ls._data) == 0:
        return _InternalSelector()
    reqs = []
    for k, v in ls._data.items():
        reqs.append(Requirement(key=k, op=Operator.Equals, vals=[v]))
    return _InternalSelector(data=reqs)


def selector_from_set(ls: Set) -> Selector:
    return selector_from_validated_set(ls)


every_thing = _InternalSelector()
nothing = _NothingSelector()

def new_selector() -> Selector:
    return _InternalSelector()




class Entry(object):

    def __init__(self, key, value):
        self.key = key
        self.value = value


class Cache(object):

    def __init__(self, max_entries: int):
        self.max_entries = max_entries
        self._ll = []
        self._cache = dict()
        self.on_evicted = None

    def _move_to_front(self, entry: Entry):
        self._ll.remove(entry)
        self._ll.append(entry)

    def _push_front(self, entry):
        self._ll.append(entry)

    def remove_oldest(self):
        if len(self._ll) == 0:
            return
        entry = self._ll[0]
        self._remove_element(entry)

    def add(self, key, value):
        if key in self._cache:
            self._move_to_front(self._cache[key])
            self._cache[key].value = value
            return
        entry = Entry(key, value)
        self._push_front(entry)
        self._cache[key] = entry
        if 0 < self.max_entries < len(self._ll):
            self.remove_oldest()

    def get(self, key):
        if key in self._cache:
            self._move_to_front(self._cache[key])
            return self._cache[key].value, True
        return None, False

    def _remove_element(self, entry: Entry):
        self._ll.remove(entry)
        del self._cache[entry.key]
        if self.on_evicted is not None:
            self.on_evicted(entry.key, entry.value)

    def remove(self, key):
        if key in self._cache:
            self._remove_element(self._cache[key])

    def __len__(self):
        return len(self._ll)

    def clear(self):
        if self.on_evicted is not None:
            for k, entry in self._cache.items():
                self.on_evicted(entry.key, entry.value)
        self._ll.clear()
        self._cache.clear()



import unittest


class _WaitFor(object):

    def __init__(self, data, ready_at):
        self.data = data
        self.ready_at = ready_at
        self.index = 0


class _PriorityQueue(object):

    def __init__(self, size=1000):
        self._data = [_WaitFor(None, 0)] * size
        self._size = size
        self._len = 0

    def is_full(self):
        return self._len == self._size

    def is_empty(self):
        return self._len == 0

    def add(self, x: _WaitFor):
        if self.is_full():
            raise Exception("is full")
        i = self._len
        self._len += 1
        while i > 0:
            p = int((i-1)/2)
            if self._data[p].ready_at <= x.ready_at:
                break
            self._data[i] = self._data[p]
            i = p
        self._data[i] = x

    def get(self) -> _WaitFor|None:
        if self.is_empty():
            return None
        res = self._data[0]
        i = 0
        x = self._data[self._len-1]
        self._len -= 1
        while 2 * i + 1 < self._len:
            a = 2 * i + 1
            b = 2 * i + 2
            if b < self._len and self._data[b].ready_at < self._data[a].ready_at:
                a = b
            if x.ready_at < self._data[a].ready_at:
                break
            self._data[i] = self._data[a]
            i = a
        self._data[i] = x
        return res


class UtilTest(unittest.TestCase):

    def test1(self):
        p = _PriorityQueue()
        p.add(_WaitFor(1, 1))
        p.add(_WaitFor(2, 2))
        p.add(_WaitFor(10, 10))
        p.add(_WaitFor(8, 8))

        print(p.get().ready_at)
        print(p.get().ready_at)
        print(p.get().ready_at)
        print(p.get().ready_at)


